package ac.ed.lurg.country.gams;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import com.gams.api.GAMSDatabase;
import com.gams.api.GAMSGlobals;
import com.gams.api.GAMSJob;
import com.gams.api.GAMSOptions;
import com.gams.api.GAMSParameter;
import com.gams.api.GAMSParameterRecord;
import com.gams.api.GAMSVariable;
import com.gams.api.GAMSVariableRecord;
import com.gams.api.GAMSWorkspace;
import com.gams.api.GAMSWorkspaceInfo;

import ac.ed.lurg.ModelConfig;
import ac.ed.lurg.types.CropType;
import ac.ed.lurg.utils.LogWriter;

public class GamsDemandOptimiser {
	private static final boolean DEBUG = true;

	private double income;


	public static void main(String[] args)  {
		BufferedWriter demandWriter = null;

		try {
			demandWriter = new BufferedWriter(new FileWriter(ModelConfig.OUTPUT_DIR + "/demand.out", false));

			StringBuffer sbHeader = new StringBuffer("income");
			for (CropType crop : CropType.getImportedTypes())
				sbHeader.append("," + crop.getGamsName());

			demandWriter.write(sbHeader.toString());
			demandWriter.newLine();

			for (int i=0; i<50; i++) {
				double income = 20.0 + i * 2.0;
				GamsDemandOptimiser theModel = new GamsDemandOptimiser(income);
				Map<CropType, Double> results = theModel.run();

				StringBuffer sbData = new StringBuffer(String.format("%.6f", income));
				for (CropType crop : CropType.getImportedTypes()) {
					Double d = results.get(crop);
					sbData.append(String.format(",%.6f", d));
				}

				demandWriter.write(sbData.toString());
				demandWriter.newLine();
			}
		}
		catch (IOException e) {
			LogWriter.print(e);
		}
		finally {
			if (demandWriter != null) {
				try {
					demandWriter.close();
				} 
				catch (IOException e) {
					LogWriter.print(e);
				}
			}

		}
	}


	public GamsDemandOptimiser(double income) {
		this.income = income;
	}

	public Map<CropType, Double> run() {
		File workingDirectory = new File(ModelConfig.TEMP_DIR, "GamsTmp");
		workingDirectory.mkdir();

		GAMSWorkspaceInfo  wsInfo  = new GAMSWorkspaceInfo();
		wsInfo.setWorkingDirectory(workingDirectory.getAbsolutePath());
		//	wsInfo.setDebugLevel(DebugLevel.VERBOSE);

		GAMSWorkspace ws = new GAMSWorkspace(wsInfo);
		GAMSDatabase inDB = ws.addDatabase();

		setupInDB(inDB);

		GAMSJob gamsJob = ws.addJobFromFile(ModelConfig.BASE_DIR + File.separator + "GAMS/demand.gms");
		GAMSOptions opt = ws.addOptions();
		//		opt.setAllModelTypes("conopt");
		opt.defines("demandgdx", inDB.getName());

		long startTime = System.currentTimeMillis();
		gamsJob.run(opt, inDB);		
		LogWriter.println("Took " + (System.currentTimeMillis() - startTime) + " ms to run");

		return handleResults(gamsJob.OutDB());
	}

	private void setupInDB(GAMSDatabase inDB) {
		addScalar(inDB, "income", income, 3);
	}

	private void addScalar(GAMSDatabase gamsDb, String recordName, double val, int places) {
		GAMSParameter param = gamsDb.addParameter(recordName, 0);
		setGamsParamValue(param.addRecord(), val, places);
		if (DEBUG) LogWriter.println(recordName + ": " + val);
	}

	private void setGamsParamValue(GAMSParameterRecord param, double val, int places) {
		double dOut = Math.round(val * Math.pow(10,places)) / Math.pow(10,places);
		param.setValue(dOut);
	}

	private Map<CropType, Double> handleResults(GAMSDatabase outDB) {
		int modelStatus = (int) outDB.getParameter("ms").findRecord().getValue();
		LogWriter.println(String.format("\nModelstatus %s, Solvestatus %s", 
				GAMSGlobals.ModelStat.lookup( modelStatus ),
				GAMSGlobals.SolveStat.lookup((int) outDB.getParameter("ss").findRecord().getValue()) ));

		GAMSVariable varConsump = outDB.getVariable("consumption");

		double consumpValue;
		Map<CropType, Double> consumptionData = new HashMap<CropType, Double>();

		for (GAMSVariableRecord rec : varConsump) {
			String itemName = rec.getKeys()[0];
			consumpValue = rec.getLevel();

			CropType cropType = CropType.getForGamsName(itemName);

			consumptionData.put(cropType, consumpValue);
			if (DEBUG) LogWriter.println(String.format("\n%s:\tconsumption= %.3f", itemName, consumpValue)); 
		}

		return consumptionData;
	}
}
