package ac.ed.lurg.demand;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import ac.ed.lurg.ModelConfig;
import ac.ed.lurg.country.CountryManager;
import ac.ed.lurg.country.SingleCountry;
import ac.ed.lurg.types.CommodityType;
import ac.ed.lurg.utils.Interpolator;
import ac.ed.lurg.utils.LazyHashMap;
import ac.ed.lurg.utils.LogWriter;
import ac.ed.lurg.utils.StringTabularReader;

public class BioenergyDemandManager {
	private static final int COUNTRY_COL = 0; 
	private static final int COMMODITY_COL = 2; 
	private static final int OTHER_COL = 3; 
	
	private Map<SingleCountry, Map<CommodityType, Double>> bioenergyBaseData;
	private Map<Integer, Double> firstGenGlobalMass; // a factor
	private Map<Integer, Double> secondGenGlobalMass; // in Mt DM
	
	public BioenergyDemandManager() {
		readBioenergyBaseData();
		readBioenergyFutureData();
	}
	
	public double getFirstGenBioenergyDemand(SingleCountry country, int year, CommodityType commodity) {
		
		if (bioenergyBaseData != null && bioenergyBaseData.containsKey(country)) {
			Double d = bioenergyBaseData.get(country).get(commodity);
			double bioenergyDemandAdj = 0;
			
			if (ModelConfig.USE_BIOENERGY_TRAJECTORY) {
				bioenergyDemandAdj = getGlobalBioenergyDemand(firstGenGlobalMass, year) / firstGenGlobalMass.get(ModelConfig.BASE_YEAR);
			}
			else {
				int yearsOfChange = Math.min(ModelConfig.BIOENERGY_CHANGE_END_YEAR - ModelConfig.BIOENERGY_CHANGE_START_YEAR, year - ModelConfig.BIOENERGY_CHANGE_START_YEAR);
				bioenergyDemandAdj = yearsOfChange > 0 ?  (1.0 +  yearsOfChange * ModelConfig.BIOENERGY_CHANGE_ANNUAL_RATE) : 1.0 ;
			}
			return d == null ? 0 : d.doubleValue() * bioenergyDemandAdj * ModelConfig.BIOENERGY_DEMAND_SHIFT;
		}
		return 0.0;
	}
	
	public double getSecondGenBioenergyDemand(int year) {
		if (ModelConfig.ENABLE_GEN2_BIOENERGY)
			return getGlobalBioenergyDemand(secondGenGlobalMass, year) * ModelConfig.BIOENERGY_DEMAND_SHIFT;
		else
			return 0.0;
	}
	
	private double getGlobalBioenergyDemand(Map<Integer, Double> beValues, int year) {
		int downYear = (year/5) * 5;
		int upYear = downYear + 5;
		Double lowerD = beValues.get(downYear);
		Double upperD = beValues.get(upYear);
		double factor = ((double)(year - downYear)) / (upYear - downYear);
		Double d = Interpolator.interpolate(lowerD, upperD, factor);
		return d;
	}

	@SuppressWarnings("serial")
	private void readBioenergyBaseData() {
		
		LazyHashMap<SingleCountry, Map<CommodityType, Double>> bioenergyMap = new LazyHashMap<SingleCountry, Map<CommodityType, Double>>() {
			protected Map<CommodityType, Double> createValue() { return new HashMap<CommodityType, Double>(); }
		};
		
		String filename = ModelConfig.BIOENERGY_1GEN_BASE_DEMAND_FILE;
		try {
			BufferedReader fitReader = new BufferedReader(new FileReader(filename)); 
			String line, countryName, commodityName;
			Double other;
			fitReader.readLine(); // read header

			while ((line=fitReader.readLine()) != null) {
				String[] tokens = line.split(",");
				
				if (tokens.length < 4)
					LogWriter.printlnError("Too few columns in " + filename + ", " + line);
				
				countryName = tokens[COUNTRY_COL];
				commodityName = tokens[COMMODITY_COL];
				other = Double.valueOf(tokens[OTHER_COL]);

				SingleCountry country = CountryManager.getForName(countryName);
				CommodityType crop = CommodityType.getForFaoName(commodityName);
				
				Map<CommodityType, Double> countryData = bioenergyMap.lazyGet(country);
				countryData.put(crop, other);
			} 
			fitReader.close(); 
		
		} catch (IOException e) {
			LogWriter.printlnError("Failed in reading commodity demand fits");
			LogWriter.print(e);
		}
		LogWriter.println("Processed " + filename + ", create " + bioenergyMap.size() + " country commodity maps values");
		
		bioenergyBaseData = bioenergyMap;
	}

	public void readBioenergyFutureData() {
		StringTabularReader beReader = new StringTabularReader(",", new String[]{"SCENARIO", "REGION", "Variable", "Year", "value"});
		beReader.read(ModelConfig.BIOENERGY_FUTURE_DEMAND_FILE);
		if (ModelConfig.USE_BIOENERGY_TRAJECTORY) firstGenGlobalMass = getGlobalBE(beReader, "Agricultural Demand|Bioenergy|1st generation");
		if (ModelConfig.ENABLE_GEN2_BIOENERGY) secondGenGlobalMass = getGlobalBE(beReader, "Agricultural Demand|Bioenergy|2nd generation");
	}

	private Map<Integer, Double> getGlobalBE(StringTabularReader beReader, String variableString) {
		Map<Integer, Double> beGlobalValues = new HashMap<Integer, Double>();
		
		Map<String, String> queryMap = new HashMap<String, String>();
		queryMap.put("SCENARIO", ModelConfig.BIOENERGY_DEMAND_SCENARIO);
		queryMap.put("REGION", "World");
		queryMap.put("Variable", variableString);
		List<Map<String, String>> rows = beReader.query(queryMap);

		for (Map<String, String> row : rows) {
			String valueS = row.get("value");
			String yearS = row.get("Year");
			Double d = Double.valueOf(valueS);
			Integer y = Integer.valueOf(yearS);
			beGlobalValues.put(y, d);
		}
		
		if (beGlobalValues.size()==0) {
			throw new RuntimeException("BioenergyDemandManager: not found values for " + variableString);
		}
		
		return beGlobalValues;
	}
}