package ac.ed.lurg.types;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Map;

import ac.ed.lurg.ModelConfig;
import ac.ed.lurg.utils.LogWriter;

public enum CropType {

	WHEAT("WheatBarleyOats", "wheat", ModelConfig.CALIB_FACTOR_WHEAT, true, false, true),
	MAIZE("MaizeMilletSorghum", "maize", ModelConfig.CALIB_FACTOR_MAIZE, true, false, true),
	RICE("Rice (Paddy Equivalent)", "rice", ModelConfig.CALIB_FACTOR_RICE, true, false, false),
	OILCROPS("Oilcrops", "oilcrops", ModelConfig.CALIB_FACTOR_OILCROPS, true, false, true),
	PULSES("Pulses", "pulses", ModelConfig.CALIB_FACTOR_PULSES, true, false, false),
	STARCHY_ROOTS("Starchy Roots", "starchyRoots", ModelConfig.CALIB_FACTOR_STARCHY_ROOTS, true, false, false),
	ENERGY_CROPS("Energy crops", "energycrops", Double.NaN, true, false, false), // no calibration factor as difference depending on whether SRC or Miscanthus
	MONOGASTRICS("Monogastrics", "monogastrics", 1.0, true, true, false),
	RUMINANTS("Ruminants", "ruminants", 1.0, true, true, false),
	PASTURE("pasture", "pasture", 1.0, false, false, true);  // confusing having a land cover and a crop type.  Needed here for yields (but not used for cropland area fractions).

	private String faoName;
	private String gamsName;
	private boolean importedCrop;
	private boolean isMeat;
	private boolean isFeedCrop;
	private double calibrationFactor;
	
	private static Collection<CropType> importedCrops;

	private CropType (String faoName, String gamsName, double calibrationFactor, boolean importedCrop, boolean isMeat, boolean isFeedCrop) {
		this.faoName = faoName;
		this.gamsName = gamsName;
		this.calibrationFactor = calibrationFactor;
		this.importedCrop = importedCrop;
		this.isMeat = isMeat;
		this.isFeedCrop = isFeedCrop;
	}
	
	public static Collection<CropType> getCropsLessPasture() {
		Collection<CropType> comms = new HashSet<CropType>();
		
		for (CropType c : values())
			if (c.importedCrop && !c.isMeat)
				comms.add(c);
		
		return comms;
	}

	public static synchronized Collection<CropType> getImportedTypes() {
		if (importedCrops == null) {
			importedCrops = new LinkedList<CropType>(); // what consistent order
		
			for (CropType c : values())
				if (c.importedCrop)
					importedCrops.add(c);
		}
		
		return importedCrops;
	}
	
	public static Collection<CropType> getNonMeatTypes() {
		Collection<CropType> comms = new HashSet<CropType>();
		
		for (CropType c : values())
			if (!c.isMeat)
				comms.add(c);
		
		return comms;
	}
	
	public static Collection<CropType> getMeatTypes() {
		Collection<CropType> comms = new HashSet<CropType>();
		
		for (CropType c : values())
			if (c.isMeat)
				comms.add(c);
		
		return comms;
	}


	private static final Map<String, CropType> faoNameCache = new HashMap<String, CropType>();
	private static final Map<String, CropType> gamsNameCache = new HashMap<String, CropType>();
    static {
        for (CropType c : values()) {
        	faoNameCache.put(c.getFaoName(), c);
        	gamsNameCache.put(c.getGamsName(), c);
       }
    }

	public static CropType getForFaoName(String faoName) {
		CropType crop = getFromCache(faoNameCache, faoName);
		if (crop == null) {
			throw new RuntimeException("Can't find CropType for faoName: " + faoName);
		}
		return crop;
	}
	
	public static CropType getForGamsName(String gamsName) {
		return getFromCache(gamsNameCache, gamsName);
	}
	
	private static CropType getFromCache(Map<String, CropType> cache, String name) {
		CropType type = cache.get(name);
		
		if (type == null)
			LogWriter.printlnError("Can't find Item for " + name);
		
		return type;
	}
	
	public static Collection<CropType> getAllItems() {
		return gamsNameCache.values();
	}

	public String getFaoName() {
		return faoName;
	}

	public String getGamsName() {
		return gamsName;
	}
	
	public boolean isImportedCrop() {
		return importedCrop;
	}

	public double getCalibrationFactor() {
		return calibrationFactor;
	}

	public boolean isFeedCrop() {
		return isFeedCrop;
	}
}
